/*
 * addr_interval_skiplist.c
 *
 *  Created on: Sep 16, 2010
 *      Author: parkcs
 */

#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include "addr_interval_skiplist.h"

static skiplist* endpoint_printer_edge_data_ref;

void endpoint_printer(skiplist_node* x, int level) {
	printf("%p => ", x->key);
	interval_set_printer((interval_set*)x->data);
	for(int i = 0; i <= level; i++) {
		printf(" |%d: ", i);
		interval_set_printer((interval_set*)skiplist_find(
				endpoint_printer_edge_data_ref, &x->forward[i]));
	}
}

void addr_islist_print(addr_islist* sl) {
	endpoint_printer_edge_data_ref = sl->edge_data;
	skiplist_print(sl->endpoints);
}

addr_islist* new_addr_islist()
{
	addr_islist* ret = (addr_islist*)malloc(sizeof(addr_islist));

	ret->endpoints = new_addr_skiplist();
	ret->endpoints->prettyprinter = endpoint_printer;

	// in case an interval starts or ends with 0x0
	// unlikely, but prevents NPE
	//ret->endpoints->head->data = new_interval_set();

	ret->edge_data = new_addr_skiplist();
	//ret->interval_sets = new_addr_skiplist();

	return ret;
}


typedef struct nodeinfo_t {
	addr_interval* I;
	skiplist_node* start_node;
	int start_level;
	uintptr_t end_key;
	struct nodeinfo_t* next;
} nodeinfo;

/* [private helper function for addr_islist_insert_endpoint]
 * removes markers along search path for I starting from start upto end
 * Note: Remove markers on endpoints as well if included in I
 * (the markers should be added later)
 */
void addr_islist_remove_markers(nodeinfo* n, skiplist* markers)
{
	skiplist_node* x = n->start_node;

	for(int i = n->start_level; i >= 0; i--) {
		while(x->forward[i] && (uintptr_t)x->forward[i]->key < n->end_key) {
			interval_set* is = skiplist_find(markers, &(x->forward[i]));
			if(is)
				interval_set_remove(is, n->I);
			if(is_in_interval((uintptr_t)x->key, *(n->I)))
				interval_set_remove(x->data, n->I);
//			if(is_in_interval(x->forward[i]->key, *(n->I)))
//				interval_set_remove(x->forward[i]->data, n->I);
			x = x->forward[i];
		}
		// Here, x->key < n->end_key <= x->forward[i]->key
	}

	if(x->forward[0] && (uintptr_t)x->forward[0]->key == n->end_key) {
		interval_set* is = skiplist_find(markers, &(x->forward[0]));
		if(is)
			interval_set_remove(is, n->I);
		if(is_in_interval((uintptr_t)x->key, *(n->I)))
			interval_set_remove(x->data, n->I);
		if(is_in_interval((uintptr_t)x->forward[0]->key, *(n->I)))
			interval_set_remove(x->forward[0]->data, n->I);
	}
}

/* [private helper function for addr_islist_insert_endpoint]
 * adds markers along search path for I starting from start_node upto end_key
 * Note: "start_key" and "end_key" are all in "endpoints" (simplifies search)
 */
void addr_islist_add_markers(nodeinfo* n, skiplist* markers)
{
	skiplist_node* x = n->start_node;

	if(is_in_interval((uintptr_t)x->key, *(n->I)))
		interval_set_add(x->data, n->I);

	for(int i = n->start_level; i >= 0; i--) {
		while(x->forward[i] && (uintptr_t)x->forward[i]->key < n->end_key) {
			interval_set* is = skiplist_find(markers, &(x->forward[i]));
			if(!is) {
				is = new_interval_set();
				skiplist_insert(markers, &(x->forward[i]), is);
			}
			interval_set_add(is, n->I);
			interval_set_add(x->forward[i]->data, n->I);
			x = x->forward[i];
		}
		// Here, x->key < n->end_key <= x->forward[i]->key
		if(x->forward[i] && (uintptr_t)x->forward[i]->key == n->end_key) {
			interval_set* is = skiplist_find(markers, &(x->forward[i]));
			if(!is) {
				is = new_interval_set();
				skiplist_insert(markers, &(x->forward[i]), is);
			}
			interval_set_add(is, n->I);
			if(is_in_interval((uintptr_t)x->forward[i]->key, *(n->I)))
				interval_set_add(x->forward[i]->data, n->I);
			x = x->forward[i];
			break;
		}

	}

#if 0
/*	if(add_endpoints && (uintptr_t)x->key == n->I->end_addr
			&& is_right_closed(*(n->I))) {
		interval_set_add(x->data, n->I);
	}*/

	//x = x->forward[0];
	if(x->forward[0] && (uintptr_t)x->forward[0]->key == n->end_key) {
		interval_set* is = skiplist_find(markers, &(x->forward[0]));
		if(!is) {
			is = new_interval_set();
			skiplist_insert(markers, &(x->forward[0]), is);
		}
		interval_set_add(is, n->I);

		if(is_in_interval((uintptr_t)x->key, *(n->I)))
			interval_set_add(x->data, n->I);
		if(is_in_interval((uintptr_t)x->forward[0]->key, *(n->I)))
			interval_set_add(x->forward[0]->data, n->I);
	}
#endif
}

#define interval_set_foreach(s, body)	do {\
			if(s) {\
				skiplist_node* _x = (s)->head;\
				while(_x) {\
					if(x->data) {\
						addr_interval* x = _x->key;\
						body;\
					}\
					_x = _x->forward[0];\
		}	}	} while(0)

/*
void interval_set_foreach(interval_set* s,
		void (*f)(interval_set*, addr_interval const *),
		interval_set* s2)
{
	skiplist_node* x = s->head;
	while(x) {
		if(x->data) {
			f(s2, x->key);
		}
		x = x->forward[0];
	}
}
*/

/* [private helper for addr_islist_insert_endpoint]
 * returns node for endpoint max(L->key, left) and its level

nodeinfo get_max(nodeinfo* ret, skiplist_node* L, int level, void* left, int (*comparator)(void*, void*))
{
	if(comparator(left, L->key) <= 0) {
		ret->
		return;
	}

	while(ret.level >= 0) {
		while(ret.node->forward[level] &&
				((uintptr_t)ret.node->forward[level]->key < (uintptr_t)left)) {
			ret.node = ret.node->forward[level];
		}

		if(ret.node->key == left)
			return ret;
		else
			--(ret.level);
	}
	abort();
}

uintptr_t get_min(skiplist_node* R, void* right, int (*comparator)(void*, void*))
{
	if(R && comparator(R->key, right) <= 0)
		return (uintptr_t) R->key;
	return (uintptr_t) right;
}
*/

void set_max(nodeinfo* n)
{
	if(n->I->start_addr <= (uintptr_t)n->start_node->key)
		return;

	skiplist_node* x = n->start_node;
	uintptr_t left = n->I->start_addr;

	for(int i = n->start_level; i >= 0; i--) {
		while(x->forward[i] && (uintptr_t)x->forward[i]->key <= left)
			x = x->forward[i];

		if(((uintptr_t)x->key) == left) {
			n->start_node = x;
			n->start_level = i;
			return;
		}
	}
	abort();
}

/* [private helper function for addr_islist_insert]
 * Adds an endpoint of an interval to the interval skip list
 * adjusts existing markers if endpoint not already in list
 */
void addr_islist_insert_endpoint(addr_islist* isl, uintptr_t key)
{
	skiplist_node* update[MAX_SKIPLIST_LEVEL];
	skiplist* sl = isl->endpoints;
	nodeinfo* affected = 0;
    //int (*comparator)(void*, void*) = sl->comparator;

	skiplist_node* x = sl->head;
	for(int i = sl->level; i >=0; i--) {
		while(x->forward[i] && (uintptr_t)x->forward[i]->key < key) {
			x = x->forward[i];
		}
		update[i] = x;
	}
	x = x->forward[0];
    if(x && (uintptr_t)x->key == key) {	// endpoint already in skiplist
		return;
	}
	else {	// add endpoint into skiplist
		int level = geomrand();
		if (level > sl->level) {	// Using "Fix the dice" method of [Pugh90]
			level = ++(sl->level);
			update[level] = sl->head;
		}

		// removes intervals from update[*]
		skiplist_node* L = update[level];
		skiplist_node* R = update[level]->forward[level];
		for(int i = 0; i <= level; i++) {
			/* We're searching through the markers on the forward edge
			 * Thus, each x corresponds to an interval
			 */
			//interval_set_foreach((interval_set*)skiplist_find(isl->edge_data, &(update[i]->forward[i])),
			interval_set* markers = (interval_set*)skiplist_find(isl->edge_data,
					&(update[i]->forward[i]));
			if(markers) {
				skiplist_node* _x = markers->head;
				while(_x) {
					addr_interval* x;
					if((x = _x->key)) {
						//interval_set_add(fixed, x);
						nodeinfo* old_affected = affected;
						affected = alloca(sizeof(nodeinfo));
						affected->I = x;
						affected->start_node = L;
						affected->start_level = i;
						affected->end_key = (R && (uintptr_t)R->key <= x->end_addr) ?
								(uintptr_t)R->key : x->end_addr;
						affected->next = old_affected;
						set_max(affected);
						addr_islist_remove_markers(affected, isl->edge_data);
					}
					_x = _x->forward[0];
				}
			}
			//);
		}

		// add endpoint to skiplist
		x = sl->allocator(level);
		x->key = (void*)key;
		x->data = new_interval_set();
		for(int i = 0; i <= level; i++) {
			x->forward[i] = update[i]->forward[i];
			update[i]->forward[i] = x;
		}

		// add back intervals from update[*]
		while(affected) {
			addr_islist_add_markers(affected, isl->edge_data);
			affected = affected->next;
		}
	}

}

skiplist_node* addr_islist_insert(addr_islist* sl,
		addr_interval* key, void* data)
{
	/* This code is wrong:
	 * markers is not the cache for all intervals in isl
	skiplist_node* old;
	if( (old = skiplist_insert(sl->markers, key, data)) ) {
		// TODO: handle conflict in same memory region
		// punt to caller for now
		return old;
	}*/

	// insert key->start_addr if necessary
	addr_islist_insert_endpoint(sl, key->start_addr);

	// insert key->end_addr if necessary
	addr_islist_insert_endpoint(sl, key->end_addr);

	// add markers for [key->start_addr, key->end_addr)
	// Make sure markers on nodes are added appropriately
	nodeinfo n;
	n.I = key;
	n.start_node = sl->endpoints->head;
	n.start_level = sl->endpoints->level;
	n.end_key = key->end_addr;
	set_max(&n);
	addr_islist_add_markers(&n, sl->edge_data);

	return 0;
}

/* Returns all intervals that contain the point key
 * responsibility of caller to free interval_set using delete_interval_set()
 */
interval_set* addr_islist_overlap_point(addr_islist* sl, void* key)
{
	interval_set* ret = new_interval_set();
	int (*comp)(void*, void*) = sl->endpoints->comparator;
	skiplist_node* x = sl->endpoints->head;
	for(int i = sl->endpoints->level; i >= 0; i--) {
		while(x->forward[i] && comp(x->forward[i]->key, key) < 0) {
			x = x->forward[i];
		}
		if(x->forward[i] && comp(x->forward[i]->key, key) == 0) {
			interval_set_addall(ret, (interval_set*)x->forward[i]->data);
			break;
			//x = x->forward[i];
		}

		interval_set_addall(ret,
				(interval_set*)skiplist_find(sl->edge_data, &(x->forward[i])));
	}

#if 0
	while(x->forward[0] && comp(x->forward[0]->key, key) < 0)
		x = x->forward[0];

	if(!x->forward[0] || comp(x->forward[0]->key, key) != 0) {
		interval_set_addall(ret,
				(interval_set*)skiplist_find(sl->edge_data, &(x->forward[0])));
	}
	else {
		interval_set_addall(ret, (interval_set*)x->forward[0]->data);
	}
#endif

	return ret;
}


/* Some optimizations can be done if all intervals in islist are
 * of the form [a,b) (as in the case with memory addresses)
 * -- only need to scan for the left endpoint and add all data following
 * level 0 forward edges until right endpoint
 */
interval_set* addr_islist_overlap_interval(addr_islist* sl, addr_interval* I)
{
	interval_set* ret = new_interval_set();
	int (*comp)(void*, void*) = sl->endpoints->comparator;
	skiplist_node* x = sl->endpoints->head;

	for(int i = sl->endpoints->level; i > 0; i--) {
		// First search for left endpoint a of I=[(a,b)]
		while(x->forward[i] && comp(x->forward[i]->key, (void*)I->start_addr) < 0) {
			x = x->forward[i];
		}
		if(x->forward[i] && comp(x->forward[i]->key, (void*)I->start_addr) == 0) {
			x = x->forward[i];
			if(is_left_closed(*I)) {
				interval_set_addall(ret, (interval_set*)x->data);
			}
		}
		// At this point, x->key < a <= x->forward[i]->key, thus, all
		// intervals on forward edges of level i up to b is overlapping
		skiplist_node* y = x;
		while(y->forward[i] && comp(y->forward[i]->key, (void*)I->end_addr) < 0) {
			interval_set_addall(ret, (interval_set*)skiplist_find(
					sl->edge_data, &(y->forward[i])));
			y = y->forward[i];
		}
		interval_set_addall(ret, (interval_set*)skiplist_find(
				sl->edge_data, &(y->forward[i])));
	}

	// Special case for level 0: inclusion of endpoints dependent on type
	while(x->forward[0] && comp(x->forward[0]->key, (void*)I->start_addr) < 0) {
		x = x->forward[0];
	}
	// x->key < a <= x->forward[0]->key
	if(x->forward[0] && comp(x->forward[0]->key, (void*)I->start_addr) == 0) {
		x = x->forward[0];
		if(is_left_closed(*I))
			interval_set_addall(ret, (interval_set*)x->data);
	}

	// add all intervals on forward edges and nodes until we reach b
	while(x->forward[0] && comp(x->forward[0]->key, (void*)I->end_addr) < 0) {
		interval_set_addall(ret, (interval_set*)skiplist_find(
				sl->edge_data, &(x->forward[0])));
		interval_set_addall(ret, (interval_set*)x->forward[0]->data);
		x = x->forward[0];
	}
	// x->key < b <= x->forward[0]->key
	interval_set_addall(ret, (interval_set*)skiplist_find(
			sl->edge_data, &(x->forward[0])));
	if(x->forward[0] && comp(x->forward[0]->key, (void*)I->end_addr) == 0 && is_right_closed(*I))
		interval_set_addall(ret, (interval_set*)x->forward[0]->data);

	return ret;
}

