/*
 * interval_set.h
 *
 *  Created on: Sep 20, 2010
 *      Author: parkcs
 */

#ifndef INTERVAL_SET_H_
#define INTERVAL_SET_H_

#include <stdbool.h>
#include <stdint.h>
#include "skiplist.h"

enum interval_type {
	OPEN, LEFT_OPEN, RIGHT_OPEN, CLOSED
};

#define is_left_closed(i)	((i).type & 2)
#define is_right_closed(i)	((i).type & 1)
#define is_in_interval(p, i)	((p == (i).start_addr && is_left_closed(i))\
		|| ((i).start_addr < (p) && (p) < (i).end_addr)\
		|| ((p) == (i).end_addr && is_right_closed(i)))
#define overlaps_interval(i1, i2)	\
		(((i1).start_addr < (i2).end_addr && (i2).start_addr < (i1).end_addr)\
		|| ((i1).end_addr == (i2).start_addr && is_right_closed(i1) && is_left_closed(i2))\
		|| ((i1).start_addr == (i2).end_addr && is_left_closed(i1) && is_right_closed(i2)))

typedef struct addr_interval_t {
	uintptr_t start_addr;
	uintptr_t end_addr;
	enum interval_type type;
} addr_interval;

typedef skiplist interval_set;

uintptr_t interval_set_hashcode(interval_set* s);
bool interval_set_add(interval_set* s, addr_interval* i);
bool interval_set_addall(interval_set* s, interval_set* is);
bool interval_set_remove(interval_set* s, addr_interval* i);
interval_set* new_interval_set();
void delete_interval_set(interval_set* s);
int interval_set_size(interval_set* s);
void interval_printer(addr_interval* i);
void interval_set_printer(interval_set* s);




#endif /* INTERVAL_SET_H_ */
