/*
 * skiplist.c
 *
 *  Created on: Sep 13, 2010
 *      Author: parkcs
 */

#include <stdlib.h>
#include <stdio.h>
#include "skiplist.h"



/* return data associated with key or NULL if not found */
void* skiplist_find(skiplist* sl, void* key)
{
	skiplist_node* x = sl->head;
	for(int i = sl->level; i >=0; i--) {
		while(x->forward[i] && sl->comparator(x->forward[i]->key, key) < 0) {
			x = x->forward[i];
		}
	}
	x = x->forward[0];
	if (x && sl->comparator(x->key, key) == 0)
		return x->data;
	else
		return NULL;
}

skiplist_node* skiplist_insert(skiplist* sl, void* key, void* data)
{
	if(!sl)
		return 0;

	skiplist_node* update[MAX_SKIPLIST_LEVEL];
	skiplist_node* x = sl->head;

	for(int i = sl->level; i >=0; i--) {
		while(x->forward[i] && sl->comparator(x->forward[i]->key, key) < 0) {
			x = x->forward[i];
		}
		update[i] = x;
	}
	x = x->forward[0];
	if (x && sl->comparator(x->key, key) == 0) {
		return x;
	}
	else {
		int level = geomrand();
		if (level > sl->level) {	// Using "Fix the dice" method of [Pugh90]
			level = ++sl->level;
			update[level] = sl->head;
		}
		x = sl->allocator(level);
		x->key = key;
		x->data = data;
		for(int i = 0; i <= level; i++) {
			x->forward[i] = update[i]->forward[i];
			update[i]->forward[i] = x;
		}

		return 0;
	}
}

void skiplist_print(skiplist* sl)
{
	char links[] = "|||||||||||||||||||||||||||||||";
	skiplist_node* prevs[MAX_SKIPLIST_LEVEL] = {0};
	printf("Skiplist with %d levels\n", 1+sl->level);

	for(int i=0; i <= sl->level; i++)
		prevs[i] = sl->head;

	skiplist_node* x = sl->head;
	while(x) {
		int i = 0;
		for(; i <= sl->level; i++) {
			if(prevs[i]->forward[i] == x) {
				printf("*");
				prevs[i] = x;
			}
			else {
				printf("%s", &links[MAX_SKIPLIST_LEVEL-sl->level+i-1]);
				break;
			}
		}
		if(sl->prettyprinter) {
			printf("=> ");
			sl->prettyprinter(x, i-1);
			printf("\n");
		}
		else {
			printf(" <key:%p, data:%p>\n", x->key, x->data);
		}
		x = x->forward[0];
	}
}
