#include "upct_list.h"

void list_init(shared list_elem * head) {
    head->data = NULL;
    head->next = NULL;
}

void list_add(shared list_elem * head, shared void * data) {
    assert(head != NULL);
    if (head->data == NULL) {
        head->data = data;
    } else {
        shared list_elem * current = head;
        shared list_elem * save = head->next;
        shared list_elem * new_element = 
            (shared list_elem *) upc_alloc(sizeof(list_elem));
        new_element->data = data;
        new_element->next = save;
        head->next = new_element;
    }
}

int list_size(shared list_elem * head) {
    if (head->data == NULL) {
        return 0;
    } else {
        int length = 0;
        shared list_elem * current = head;
        while (current != NULL) {
            length++;
            current = current->next;
        }
        return length;
    }
}

void list_print(shared list_elem * head, void (*print_func)(shared void *)) {
    printf("List (size: %d) [", list_size(head));
    while (head != NULL && head->data != NULL) {
        print_func(head->data);
        head = head->next;
    }
    printf("]");
}

shared list_elem * list_get_next(shared list_elem * el) {
    return el->next;
}

shared void * list_get_data(shared list_elem * el) {
    return el->data;
}


int list_contains(shared list_elem * list, 
                      shared void * target,
                      int (*comp_lt) (shared void *, shared void *)) {
    shared list_elem * current = list;
    while (current != NULL) {
        shared void * my_data = list_get_data(current);
        int comparison_lt = comp_lt(my_data, target);
        int comparison_gt = comp_lt(target, my_data);
        if ((! comparison_lt) && (! comparison_gt)) {
            return FOUND;
        }
        current = list_get_next(current);
    }
    return NOT_FOUND;
}

int list_intersection(shared list_elem * first,
                          shared list_elem * second, 
                          int (*comp_lt) (shared void *, shared void *)) {
    if ((list_size(first) == 0) || (list_size(second) == 0)) {
        return EMPTY;
    }
    shared list_elem * current = first;
    while (current != NULL) {
        shared void * tmp_data = list_get_data(current);
        if (list_contains(second, tmp_data, comp_lt)) {
            return NONEMPTY;
        }
        current = list_get_next(current);
    }
    return EMPTY;
}


