/*
 * upct_misc.c -- implementation of arch dependent functions
 *
 */

#include "upct_types.h"
#include "upct_list.h"
#include <stdio.h>

#define RACE_INFO_FILENAME	"upct.race"
#define MAX_FILENAME_SIZE	256
#define MAX_LINE_SIZE	1024

/* Copy local string to shared 
 * src must be null terminated */

/*
void ls_strcpy(const char* src, shared [] const char* dst) {
    int n = strlen(src) + 1;
    shared [] char* tmp = (shared [] char*) upc_alloc(n);
    upc_memput(tmp, src, n);
    dst = tmp;
}
*/

/* Copy shared string to local */
/*
void sl_strcpy(shared [] const char* src, const char* dst) {
    int n = strlen(src) + 1;
    shared [] char* tmp = (shared [] char*) upc_alloc(n);
    upc_memput(tmp, src, n);
    dst = tmp;
}
*/

void print_lock_list(shared void * ls) {
    shared lock_addr * mylock = (shared lock_addr *) ls;
    printf("%lx",(*mylock));
}

void print_race_info(shared race_info_t * info) {
    printf("Raceinfo(Addr:%lx, ", info->addr);
    printf("Thr:%d, ", info->thread);
    printf("Phase:%d, ", info->phase);
    printf("Access:%s, ", info->access == READ ? "R" : "W");
    printf("LocksetSize:%d, ", list_size(info->locklist));
    printf("Lockset:");
    list_print(info->locklist, &print_lock_list);
    printf(", ");
    printf("File:%s, ", info->file);
    printf("Expr:%s, ", info->expr);
    printf("Line:%d, ", info->line);
    printf("RaceReported:%s, ", info->race_reported == REPORTED ? "Y": "N");
    printf("Owned:%s", info->owned == TRUE ? "Y": "N");
    printf(")");
}

void print_race_info_ns(race_info_t * info) {
    printf("Raceinfo(Addr:%lx, ", info->addr);
    printf("Thr:%d, ", info->thread);
    printf("Phase:%d, ", info->phase);
    printf("Access:%s, ", info->access == READ ? "R" : "W");
    printf("LocksetSize:%d, ", list_size(info->locklist));
    printf("Lockset:");
    list_print(info->locklist, &print_lock_list);
    printf(", ");
    printf("File:%s, ", info->file);
    printf("Expr:%s, ", info->expr);
    printf("Line:%d, ", info->line);
    printf("RaceReported:%s", info->race_reported == REPORTED ? "Y": "N");
    printf("Owned:%s", info->owned == TRUE ? "Y": "N");
    printf(")");
}

char fname[MAX_FILENAME_SIZE];

void write_race_info(shared race_info_t *info, FILE *fp) {
    fprintf(fp, "%ld\n", (long) info->addr);
    fprintf(fp, "%d\n", info->thread);
    fprintf(fp, "%d\n", info->phase);
    fprintf(fp, "%d\n", info->access);
    // Skipping lockset
    fprintf(fp, "%s\n", info->file);
    fprintf(fp, "%s\n", info->expr);
    fprintf(fp, "%d\n", info->line);
    fprintf(fp, "%d\n", info->owned);
}

void read_race_info(race_info_t *info, FILE* fp) {
    char buf[MAX_LINE_SIZE];
    int n;
    fgets(buf, MAX_LINE_SIZE, fp);
    info->addr = atol(buf);
    //printf("ADDR = %p\n", info->addr);	// debug

    fgets(buf, MAX_LINE_SIZE, fp);
    info->thread = atoi(buf);
    //printf("THREAD = %d\n", info->thread);	// debug

    fgets(buf, MAX_LINE_SIZE, fp);
    info->phase = atoi(buf);

    fgets(buf, MAX_LINE_SIZE, fp);
    info->access = atoi(buf);
    //printf("ACCESS = %s\n", info->access == READ ? "R" : "W");	// debug

    info->locklist = 0;	// Skipping lockset
    //locklist_print(info->locklist);	// debug

    fgets(buf, MAX_LINE_SIZE, fp);
    n = strlen(buf);
    buf[n-1] = 0;	// remove newline
    ////printf("FILE(local) = %s\n", buf);	// debug
    char * tmp = (char*) malloc(n);
    strncpy(tmp, buf, n);
    info->file = tmp;
    /*
    //= (shared [] char*) upc_alloc(n);
    //upc_memput(info->file, buf, n);
    ls_strcpy(buf, info->file);
    */
    //printf("FILE = %s\n", info->file);	// debug

    fgets(buf, MAX_LINE_SIZE, fp);
    n = strlen(buf);
    buf[n-1] = 0;	// remove newline
    
    tmp = (char*) malloc(n);
    strncpy(tmp, buf, n);
    info->expr = tmp;

    /*
    //info->expr =  (shared [] char*) upc_alloc(n);
    //upc_memput(info->expr, buf, n);
    ls_strcpy(buf, info->expr);
    */
    //printf("EXPR = %s\n", info->expr);	// debug

    fgets(buf, MAX_LINE_SIZE, fp);
    info->line = atoi(buf);
    //printf("LINE = %d\n", info->line);	// debug

    fgets(buf, MAX_LINE_SIZE, fp);
    info->owned = atoi(buf);
    printf("OWNED = %s\n", (info->owned == TRUE) ? "Y" : "N");	// debug
}

void write_race_pair(shared race_info_t *i1, 
        shared race_info_t *i2, int id) {

    sprintf(fname, "%s.%d", RACE_INFO_FILENAME, id);
    FILE* fp = fopen(fname, "w");
    write_race_info(i1, fp);
    write_race_info(i2, fp);
    fclose(fp);
}

void read_race_pair(race_info_t *i1, 
        race_info_t *i2, int id) {

    sprintf(fname, "%s.%d", RACE_INFO_FILENAME, id);
    FILE* fp = fopen(fname, "r");
    if(fp) {
        read_race_info(i1, fp);
        read_race_info(i2, fp);
        fclose(fp);
    }
    else {
        fprintf(stderr, "UPCT: cannot read race pair file %s\n", fname);
        bupc_exit(-1);
    }
}
