#include "upct_multiset.h"

void multiset_elem_print(shared multiset_elem *, void (*)(shared void *));
shared multiset_elem * multiset_allocate_element();
void multiset_initialize_element(shared multiset_elem *, shared void *);


void multiset_init(shared multiset_t * ms, 
                   int (*comp_lt)(shared void *, shared void *)) {
    ms->root = NULL;
    ms->comp_lt = comp_lt;
}

shared multiset_elem * multiset_allocate_element() {
    shared multiset_elem * new_el;
    new_el = (shared multiset_elem *) upc_alloc(sizeof(multiset_elem));
    shared list_elem * new_list;
    new_list = (shared list_elem *) upc_alloc(sizeof(list_elem));
    new_el->list = new_list;
    return new_el;
}

void multiset_initialize_element(shared multiset_elem* el, shared void* data) {
    list_init(el->list);
    list_add(el->list, data);
    el->valid = VALID;
    el->left = NULL;
    el->right = NULL;
}



void multiset_add(shared multiset_t * ms, shared void * data) {
    if (ms->root == NULL) {
        shared multiset_elem * new_el = multiset_allocate_element();
        multiset_initialize_element(new_el, data);
        ms->root = new_el;
    } else {
        shared multiset_elem * current = ms->root;
        while (1) {
            shared void * head = list_get_data(current->list);
            if (ms->comp_lt(data, head)) {
                if (current->left == NULL) {
                    shared multiset_elem * new_el = multiset_allocate_element();
                    multiset_initialize_element(new_el, data);
                    current->left = new_el;
                    return;
                }
                current = current->left;
            } else if (ms->comp_lt(head, data)) {
                if (current->right == NULL) {
                    shared multiset_elem * new_el = multiset_allocate_element();
                    multiset_initialize_element(new_el, data);
                    current->right = new_el;
                    return;
                }
                current = current->right;
            } else {
                if (current->valid == INVALID) {
                    current->valid = VALID;
                }
                list_add(current->list, data);
                return;
            }
        }
    }
}


void multiset_remove(shared multiset_t * ms, shared void * target) {
    //TODO: implement
    assert(FALSE);
    return;
}

int multiset_elem_size(shared multiset_elem * el) {
    int left_size = 0;
    int right_size = 0;
    if (el->left != NULL) {
        left_size = multiset_elem_size(el->left);
    }
    if (el->right != NULL) {
        right_size = multiset_elem_size(el->right);
    }
    
    if (el->valid == VALID) {
        return (list_size(el->list) + left_size + right_size);
    } else {
        return (left_size + right_size);
    }
}


int multiset_size(shared multiset_t * ms) {
    if (ms->root == NULL) {
        return 0;
    } else {
        return multiset_elem_size(ms->root);
    }
}

void multiset_elem_print(shared multiset_elem * el, 
        void (*print_func)(shared void *)) {
    if (el->left != NULL) {
        multiset_elem_print(el->left, print_func);
    }
    if (el->valid == VALID) {
        list_print(el->list, print_func);
        printf("\n\n");
    }
    if (el->right != NULL) {
        multiset_elem_print(el->right, print_func);
    }
}

void multiset_print(shared multiset_t* ms, void (*print_func) (shared void*)) {
    if (multiset_size(ms) == 0) {
        printf("[%d] Multiset (size: %d): []\n", MYTHREAD, 0);
    } else {
        printf("[%d] Multiset (size: %d): [\n\n", MYTHREAD, multiset_size(ms));
        multiset_elem_print(ms->root, print_func);
        printf("]\n");
    }
}

shared list_elem * multiset_get(shared multiset_t * ms, shared void * target) {
    if (ms->root == NULL) {
        return NULL;
    }
    
    shared multiset_elem * current = ms->root;
    while (1) {
        shared void * head = list_get_data(current->list);
        if (ms->comp_lt(target, head)) {
            if (current->left == NULL) {
                return NULL;
            }
            current = current->left;
        } else if (ms->comp_lt(head, target)) {
            if (current->right == NULL) {
                return NULL;
            }
            current = current->right;
        } else {
            if (current->valid == INVALID) {
                return NULL;
            } else {
                return current->list;
            }
        }
    }
    return NULL;
}
