#include "upct_set.h"

void elem_print(shared set_elem *, void (*)(shared void *));
shared set_elem * set_allocate_element();
void set_initialize_element(shared set_elem *, shared void *);


void set_init(shared set_t * ls, int (*comp_lt)(shared void *, shared void *)) {
    ls->root = NULL;
    ls->comp_lt = comp_lt;
}

shared set_elem * set_allocate_element() {
    shared set_elem * new_el;
    new_el = (shared set_elem *) upc_alloc(sizeof(set_elem));
    return new_el;
}

void set_initialize_element(shared set_elem * element, shared void * data) {
    element->data = data;
    element->valid = VALID;
    element->left = NULL;
    element->right = NULL;
}

void set_add(shared set_t * ls, shared void * data) {
    if (ls->root == NULL) {
        shared set_elem * new_el = set_allocate_element();
        set_initialize_element(new_el, data);
        ls->root = new_el;
    } else {
        shared set_elem * current = ls->root;
        while (1) {
            if (ls->comp_lt(data, current->data)) {
                if (current->left == NULL) {
                    shared set_elem * new_el = set_allocate_element();
                    set_initialize_element(new_el, data);
                    current->left = new_el;
                    break;
                }
                current = current->left;
            } else if (ls->comp_lt(current->data, data)) {
                if (current->right == NULL) {
                    shared set_elem * new_el = set_allocate_element();
                    set_initialize_element(new_el, data);
                    current->right = new_el;
                    break;
                }
                current = current->right;
            } else {
                if (current->valid == INVALID) {
                    current->valid = VALID;
                }
                break;
            }
        }
    }
}


shared void * set_get(shared set_t * s, shared void * target) {
    if (s->root == NULL) {
        return NULL;
    }
    
    shared set_elem * current = s->root;
    while (1) {
        if (s->comp_lt(target, current->data)) {
            if (current->left == NULL) {
                return NULL;
            }
            current = current->left;
        } else if (s->comp_lt(current->data, target)) {
            if (current->right == NULL) {
                return NULL;
            }
            current = current->right;
        } else {
            if (current->valid = INVALID) {
                return NULL;
            } else {
                return current->data;
            }
        }
    }
    return NULL;
}

void set_remove(shared set_t * ls, shared void * target) {
    if (ls->root == NULL) {
        return;
    }
    shared set_elem * current = ls->root;
    while (1) {
        if (ls->comp_lt(target, current->data)) {
            if (current->left == NULL) {
                return;
            }
            current = current->left;
        } else if (ls->comp_lt(current->data, target)) {
            if (current->right == NULL) {
                return;
            }
            current = current->right;
        } else {
            current->valid = INVALID;
            return;
        }
    }
}

int elem_size(shared set_elem * element) {
    int left_size = 0;
    int right_size = 0;
    if (element->left != NULL) {
        left_size = elem_size(element->left);
    }
    if (element->right != NULL) {
        right_size = elem_size(element->right);
    }
    
    if (element->valid == VALID) {
        return (1 + left_size + right_size);
    } else {
        return (left_size + right_size);
    }
}


int set_size(shared set_t * ls) {
    if (ls->root == NULL) {
        return 0;
    } else {
        return elem_size(ls->root);
    }
}

void elem_print(shared set_elem * element, 
        void (*print_func)(shared void *)) {
    if (element->left != NULL) {
        elem_print(element->left, print_func);
    }
    if (element->valid == VALID) {
        printf("\t");
        print_func(element->data);
        printf(",\n");
    }
    if (element->right != NULL) {
        elem_print(element->right, print_func);
    }
}

void set_print(shared set_t * ls, void (*print_func) (shared void *)) {
    if (set_size(ls) == 0) {
        printf("[%d] Set (size: %d): []\n", MYTHREAD, set_size(ls));
    } else {
        printf("[%d] Set (size: %d): [\n", MYTHREAD, set_size(ls));
        elem_print(ls->root, print_func);
        printf("]\n");
    }
}

void elem_list(shared set_elem * element, shared list_elem * list) {
    if (element->left != NULL) {
        elem_list(element->left, list);
    }
    if (element->valid == VALID) {
        shared lock_addr * addr; 
        addr = (shared lock_addr *) upc_alloc(sizeof(lock_addr));
        shared lock_addr * my_data = (shared lock_addr *) element->data;
        (*addr) = (*my_data);
        list_add(list, (shared void *) addr);
    }
    if (element->right != NULL) {
        elem_list(element->right, list);
    }

}

shared list_elem * lockset_getlist(shared set_t * ls) {
    shared list_elem * locklist;
    locklist = (shared list_elem *) upc_alloc(sizeof(list_elem));
    list_init(locklist);
    elem_list(ls->root, locklist);
    return locklist;
}
/*
void elem_print(shared lockset_elem * element) {
    if (element->left != NULL) {
        elem_print(element->left);
    }
    if (element->valid == VALID) {
        printf("%p, ", element->lock);
    }
    if (element->right != NULL) {
        elem_print(element->right);
    }
}

void lockset_print(shared lockset_t * ls) {
    if (ls->root == NULL) {
        printf("[%d] Lockset(size: %d): []\n", MYTHREAD, lockset_size(ls));
    } else {
        printf("[%d] Lockset(size: %d): [", MYTHREAD, lockset_size(ls));
        elem_print(ls->root);
        printf("]\n");
    }
}



void elem_list(shared lockset_elem * element, shared list_elem * list) {
    if (element->left != NULL) {
        elem_list(element->left, list);
    }
    if (element->valid == VALID) {
        shared lock_addr * addr; 
        addr = (shared lock_addr *) upc_alloc(sizeof(lock_addr));
        (*addr) = element->lock;
        list_add(list, (shared void *) addr);
    }
    if (element->right != NULL) {
        elem_list(element->right, list);
    }

}

shared list_elem * lockset_getlist(shared lockset_t * ls) {
    shared list_elem * locklist;
    locklist = (shared list_elem *) upc_alloc(sizeof(list_elem));
    list_init(locklist);
    elem_list(ls->root, locklist);
    return locklist;
}
*/
