/*
 * upct_types.h - macros, typedefs, structs for upc thrille
 *
 */

#ifndef UPCT_TYPES_H
#define UPCT_TYPES_H

#include <stdio.h>
#include <upc.h>
#include <assert.h>

#define TRACE_FILENAME	"thrille.trace"
#define TRACE_FILENAME_SIZE	256	

#define THRILLE_BEGIN_FUNC(x)	if (thrille_info.inside_inst) \
                                    return x;\
                                    thrille_info.inside_inst++

#define THRILLE_END_FUNC	thrille_info.inside_inst--
#define THRILLE_LOCK		upcr_lock(thrille_lock)
#define THRILLE_UNLOCK		upcr_unlock(thrille_lock)

#define MAX_SEM_TRIES	500
#define SEM_WAIT_USEC   1000	

#define REPORTED 1
#define NOT_REPORTED 0

#define RACE 1
#define NO_RACE -1
#define BAD_RACE -2

#define EMPTY 0
#define NONEMPTY 1

#define VALID 1
#define INVALID 0

#define FOUND 1
#define NOT_FOUND 0

#define FALSE 0
#define TRUE 1

typedef long lock_addr;
typedef long mem_addr;

enum _AccessType;

struct _race_info_t;
struct _race_target_t;
struct _thrille_info_t;
struct _list_elem;

typedef enum _AccessType AccessType;

typedef struct _race_info_t race_info_t;
typedef struct _race_target_t race_target_t;
typedef struct _thrille_info_t thrille_info_t;
typedef struct _list_elem list_elem;

typedef bupc_sem_t* ptr_sem;

enum _AccessType {
    READ, 
    WRITE 
};

struct _race_info_t {
    mem_addr addr;
    int thread;
    enum _AccessType access;
    shared list_elem * locklist; 
    const char* file;
    const char* expr;
    int line;
    int race_reported;
    int phase;
    int owned;
};

struct _lockset_t;
struct _thrille_info_t {
    int inside_inst;
    shared struct _set_t * ls;
    int phase;
};

typedef struct _mem_range_t {
    uint64_t start;
    uint64_t end;
} mem_range;

/* Change this struct to a more generic bit-field for flexibity
 * in number of locks allowed / locks held in program
 */ 
typedef struct _lock_set_t {
    uint32_t idx;
    uint32_t bits;
} lock_set;

/* Minimum data required to send off to other threads for computation
 * of potential races -- compact as much as possible
 */
struct _race_data_t {
    mem_range   addr;
    lock_set    ls;
    enum _AccessType access;
};

void print_race_info(shared race_info_t *);
void print_race_info_ns(race_info_t *);
void write_race_pair(shared race_info_t *, shared race_info_t *, int);
void read_race_pair(race_info_t *, race_info_t *, int);

#endif // UPCT_TYPES_H
